pub(crate) use unnecessary_call_around_sorted::*;
pub(crate) use unnecessary_collection_call::*;
pub(crate) use unnecessary_comprehension::*;
pub(crate) use unnecessary_comprehension_in_call::*;
pub(crate) use unnecessary_dict_comprehension_for_iterable::*;
pub(crate) use unnecessary_double_cast_or_process::*;
pub(crate) use unnecessary_generator_dict::*;
pub(crate) use unnecessary_generator_list::*;
pub(crate) use unnecessary_generator_set::*;
pub(crate) use unnecessary_list_call::*;
pub(crate) use unnecessary_list_comprehension_dict::*;
pub(crate) use unnecessary_list_comprehension_set::*;
pub(crate) use unnecessary_literal_dict::*;
pub(crate) use unnecessary_literal_set::*;
pub(crate) use unnecessary_literal_within_dict_call::*;
pub(crate) use unnecessary_literal_within_list_call::*;
pub(crate) use unnecessary_literal_within_tuple_call::*;
pub(crate) use unnecessary_map::*;
pub(crate) use unnecessary_subscript_reversal::*;

mod helpers;
mod unnecessary_call_around_sorted;
mod unnecessary_collection_call;
mod unnecessary_comprehension;
mod unnecessary_comprehension_in_call;
mod unnecessary_dict_comprehension_for_iterable;
mod unnecessary_double_cast_or_process;
mod unnecessary_generator_dict;
mod unnecessary_generator_list;
mod unnecessary_generator_set;
mod unnecessary_list_call;
mod unnecessary_list_comprehension_dict;
mod unnecessary_list_comprehension_set;
mod unnecessary_literal_dict;
mod unnecessary_literal_set;
mod unnecessary_literal_within_dict_call;
mod unnecessary_literal_within_list_call;
mod unnecessary_literal_within_tuple_call;
mod unnecessary_map;
mod unnecessary_subscript_reversal;
