def long_function_name_goes_here(
    x: Callable[List[int]]
) -> Union[List[int], float, str, bytes, Tuple[int]]:
    pass


def long_function_name_goes_here(
    x: Callable[[str, Any], int]
) -> Union[List[int], float, str, bytes, Tuple[int]]:
    pass
